/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
 *  C.cpp - demonstration of canonical class structure       *
 *                                                           *
 *  Author: Jim Fawcett                                      *
 * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */


#include "c.hpp"

//----< C::C(void) >-----------------------------------------//

//  Processing:
//  - constructs C object

C::C(void) {

  if ((s = new char[1]) == NULL) {
                          // if use [] for any constructor,
                          // use for all
    cerr << "exhausted dynamic memory\n";
    exit(1);
  }
  *s = '\0';              // define null string
}

//----< C::C(C& c) >-----------------------------------------//

//  Processing:
//  - copy constructs C object from an existing object

C::C(C& c) {

  if ((s = new char[strlen(c.s)+1]) == NULL) {
    cerr << "exhausted dynamic memory\n";
    exit(1);
  }
  strcpy(s,c.s);          // copy c's data
}

//----< C::C(T& t) >-----------------------------------------//

//  Processing:
//  - constructs C object by promoting t

C::C(T& t) {

  if ((s = new char[strlen(t)+1]) == NULL) {
    cerr << "exhausted dynamic memory\n";
    exit(1);
  }
  strcpy(s,t);
}


//----< C::~C(void) >----------------------------------------//

//  Processing:
//  - destroys C object

C::~C(void) {

  delete [] s;            // always use [] if used with new
}

//----< C& C::operator=(C& c) >------------------------------//

//  Processing:
//  - sets C object equal to object c

C& C::operator=(C& c) {

  if (*this == c)  // return if already equal (need operator==)
    return *this;
  delete [] s;            // destroy self data then same as
                          // copy constructor
  if ((s = new char[strlen(c.s)+1]) == NULL) {
    cerr << "exhausted dynamic memory\n";
    exit(1);
  }
  strcpy(s,c.s);          // copy c's data
  return *this;           // operator= should always return
                          // reference to self
}

//----< C C::operator-(C& c) >-------------------------------//

//  Processing:
//  - concatenates self behind c - this is silly, but it
//    illustrates the definition mechanics

C C::operator-(C& c) {

  int length = strlen(s) + strlen(c.s);
  char *sptr;
  if ((sptr = new char[length+1]) == NULL) {
    cerr << "exhausted dynamic memory\n";
    exit(1);
  }
  strcpy(sptr,c.s);       // copy c's data
  strcat(sptr,s);         // copy this's data
  C temp(sptr);
  return temp;            // must copy temporary before it's
                          // destroyed - that's why we return
                          // copy instead of reference
}
/**/
//----< C operator+(C& c1, C& c2) >--------------------------//

//  Processing:
//  - concatenates self in front of c - this is silly, but it
//    illustrates the definition mechanics

C operator+(C& c1, C& c2) {
                          // note no class scope operator, C::
                          // note friend only appears in declar.

  int length = strlen(c1.s) + strlen(c2.s);
  char *sptr;
  if ((sptr = new char[length+1]) == NULL) {
    cerr << "exhausted dynamic memory\n";
    exit(1);
  }
  strcpy(sptr,c1.s);      // copy c1's data
  strcat(sptr,c2.s);      // copy c2's data
  C temp(sptr);
  return temp;            // must copy temporary before it's
                          // destroyed - that's why we return
                          // copy instead of reference
}

//----< char& C::operator[](int n) >-------------------------//

//  Processing:
//  - returns reference to nth character of self, counting from 0

char& C::operator[](int n) { // returning reference allows c[3] to
                             // appear on left side of expression

  if ((n < 0) || (strlen(s) < n)) {
    cerr << "index outside string\n";
    exit(1);
  }
  return s[n];
}

//----< char& C::operator()(int n, int m) >------------------//

//  Processing:
//  - returns reference to m+nth character of self, counting from 0
//    it's silly, but illustrates that double dimensions work with
//    operator() - they don't with operator[]

char& C::operator()(int n, int m) {

  int k = m + n;
  if ((k < 0) || (strlen(s) < k)) {
    cerr << "index outside string\n";
    exit(1);
  }
  return s[k];
}
/**/
//----< ostream& operator<<(ostream& ostrm, C& c) >----------//

//  Processing:
//  - copies c to stream ostrm

ostream& operator<<(ostream& ostrm, C& c) {

  ostrm << c.s;
  return ostrm;
}

//----< istream& operator>>(istream& istrm, C& c) >----------//

//  Processing:
//  - copies stream istrm to object c

istream& operator>>(istream& istrm, C& c) {

  const MAXCHARS = 512;
  char* buffer;           // allocate big buffer
  if ((buffer = new char[MAXCHARS+1]) == NULL) {
    cerr << "dynamic memory exhausted\n";
    exit(1);
  }
  istrm.getline(buffer,MAXCHARS);
                          // will truncate if larger than MAXCHARS
  delete [] c.s;          // delete this's string pointer
  int length = strlen(buffer);
  if (length > MAXCHARS)
    length = MAXCHARS;
                          // now, get buffer of correct size
  if ((c.s = new char[length+1]) == NULL) {
    cerr << "dynamic memory exhausted\n";
    exit(1);
  }
  strncpy(c.s,buffer,length);
  c.s[length] = '\0';
  delete [] buffer;
  return istrm;
}

//----< T& C::fun1(T& t) >-----------------------------------//

//  Processing:
//  - displays t string and c string, silly but illustrates ...

T& C::fun1(T& t) {        // returning reference ok if it points
                          // to existing object

  cout << t << " " << s << "\n";
  return t;               // returns input - don't ask why
}

//----< T& fun2(C& c, T& t) >--------------------------------//

//  Processing:
//  - displays c string and t string, silly but illustrates ...

T& fun2(C& c, T& t) {     // returning reference ok if it points
                          // to existing object

  cout << c.s << " " << t << "\n";
  return t;
}
